#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# SAT plugin for Jingle File Transfer (XEP-0231)
# Copyright (C) 2009-2018 Jérôme Poisson (goffi@goffi.org)

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sat.core.i18n import _
from sat.core.constants import Const as C
from sat.core import exceptions
from sat.core.log import getLogger
log = getLogger(__name__)
from sat.tools import xml_tools
from wokkel import disco, iwokkel
from zope.interface import implements
from twisted.python import failure
from twisted.words.protocols.jabber import xmlstream
from twisted.words.protocols.jabber import jid
from functools import partial
import base64


PLUGIN_INFO = {
    C.PI_NAME: "Bits of Binary",
    C.PI_IMPORT_NAME: "XEP-0231",
    C.PI_TYPE: "XEP",
    C.PI_PROTOCOLS: ["XEP-0231"],
    C.PI_DEPENDENCIES: ["XEP-0071"],
    C.PI_MAIN: "XEP_0231",
    C.PI_HANDLER: "yes",
    C.PI_DESCRIPTION: _("""Implementation of bits of binary (used for small images/files)""")
}

NS_BOB = u'urn:xmpp:bob'


class XEP_0231(object):

    def __init__(self, host):
        log.info(_(u"plugin Bits of Binary initialization"))
        self.host = host
        host.trigger.add("xhtml_post_treat", self.XHTMLTrigger)

    def dumpData(self, cache, data_elt, cid):
        """save file encoded in data_elt to cache

        @param cache(memory.cache.Cache): cache to use to store the data
        @param data_elt(domish.Element): <data> as in XEP-0231
        @param cid(unicode): content-id
        @return(unicode): full path to dumped file
        """
        # FIXME: is it needed to use a separate thread?
        #        probably not with the little data expected with BoB
        try:
            max_age = int(data_elt['max-age'])
            if max_age < 0:
                raise ValueError
        except (KeyError, ValueError):
            log.warning(u'invalid max-age found')
            max_age = None

        with cache.cacheData(
            PLUGIN_INFO[C.PI_IMPORT_NAME],
            cid,
            data_elt.getAttribute('type'),
            max_age) as f:

            file_path = f.name
            f.write(base64.b64decode(str(data_elt)))

        return file_path

    def getHandler(self, client):
        return XEP_0231_handler()

    def _requestCb(self, iq_elt, cache, cid):
        for data_elt in iq_elt.elements(NS_BOB, u'data'):
            if data_elt.getAttribute('cid') == cid:
                file_path = self.dumpData(cache, data_elt, cid)
                return file_path

        log.warning(u"invalid data stanza received, requested cid was not found:\n{iq_elt}\nrequested cid: {cid}".format(
            iq_elt = iq_elt,
            cid = cid
            ))
        raise failure.Failure(exceptions.DataError("missing data"))

    def _requestEb(self, failure_):
        """Log the error and continue errback chain"""
        log.warning(u"Can't get requested data:\n{reason}".format(reason=failure_))
        return failure_

    def requestData(self, client, to_jid, cid, cache=None):
        """Request data if we don't have it in cache

        @param to_jid(jid.JID): jid to request the data to
        @param cid(unicode): content id
        @param cache(memory.cache.Cache, None): cache to use
            client.cache will be used if None
        """
        if cache is None:
            cache = client.cache
        iq_elt = client.IQ('get')
        iq_elt['to'] = to_jid.full()
        data_elt = iq_elt.addElement((NS_BOB, 'data'))
        data_elt['cid'] = cid
        d = iq_elt.send()
        d.addCallback(self._requestCb, cache, cid)
        d.addErrback(self._requestEb)

    def _setImgEltSrc(self, path, img_elt):
        img_elt[u'src'] = u'file://{}'.format(path)

    def XHTMLTrigger(self, client, message_elt, body_elt, lang, treat_d):
        for img_elt in xml_tools.findAll(body_elt, C.NS_XHTML, u'img'):
            source = img_elt.getAttribute(u'src','')
            if source.startswith(u'cid:'):
                cid = source[4:]
                file_path = client.cache.getFilePath(cid)
                if file_path is not None:
                    # image is in cache, we change change the url
                    img_elt[u'src'] = u'file://{}'.format(file_path)
                    continue
                else:
                    # image is not in cache, is it given locally?
                    for data_elt in message_elt.elements(NS_BOB, u'data'):
                        if data_elt.getAttribute('cid') == cid:
                            file_path = self.dumpData(client.cache, data_elt, cid)
                            img_elt[u'src'] = u'file://{}'.format(file_path)
                            break
                    else:
                        # cid not found locally, we need to request it
                        # so we use the deferred
                        d = self.requestData(client, jid.JID(message_elt['from']), cid)
                        d.addCallback(partial(self._setImgEltSrc, img_elt=img_elt))
                        treat_d.addCallback(lambda dummy: d)


class XEP_0231_handler(xmlstream.XMPPHandler):
    implements(iwokkel.IDisco)

    def getDiscoInfo(self, requestor, target, nodeIdentifier=''):
        return [disco.DiscoFeature(NS_BOB)]

    def getDiscoItems(self, requestor, target, nodeIdentifier=''):
        return []
